# -*- coding: utf-8 -*-

from HTMLParser import HTMLParser
from Levenshtein import distance,ratio
import re
import string
import urllib2
import time
import sys, getopt, subprocess
import codecs
import os

#import profile


'''
Description:
    Normalizes the content of tweets:
        - Multiple Char Normalization
        - EndofSentence Normalization (EOS-N)
        - Punctuation Normalization
        - HTML Normalization
        - Laugh Normalization
        - Noisy abb Normalization
        - Special Cases Normalization
        - Spelling Correction Normalization
        - Split Strings into words (missing Space) Normalization

        - Remove Smiley and unwanted tokens
        - Remove URLs
        - Remove non-word strings containing only letters (eg. "gertherherebbbc")

        - Remove twitter tags
        - Twitter @username Normalization
        - Twitter #hashtag Normalization


Dependencies:
    - Levenshtein (Python-Modul)
    - TreeTagger (current windows version) [for abbrevation normalization]
    - langid (included)


Usage:
    Tweet file must be in the following format:
        - one tweet per Line
        OR
        - *id*\t\t*tweet* (double tap separated) "14588458        this is my tweet"

    normalize_tweets.py -i <file>
    normalize_tweets.py -h



Version:
    TweetNorm 1.0
    Python 2.7 for better performance and downward compatibility
'''


setupTimeStart = time.time()

## Boolean initialization
debug = False
fileWrite  = True
isCompareMode = False
isFullCompareMode = False
isShowUnchanged = False
isCreateSummary = True
enabledOnlineSearch = False
enabledSpellingCorrection = False
enabledSplitString = False
showUsrN = False
showHashTagN = False
isFullNorm = False
inputFileName = ""
normFlagList = [] # marks each normalized tweet
## Boolean initialization


## Read command line arguments

argList = sys.argv[1:]

try:
    opts, args = getopt.getopt(argList, "onfchudi:", ["inputfile=","debug","help","online","noSummary",
                                                     "compare","noSplit","full","showUnchanged"])
    getopt.getopt
except:
    print "normalize_tweets.py -i <inputfile>\n\n" \
          +"For more information type:\n" \
          +"normalize_tweets.py -h"
    sys.exit(2)

for opt, arg in opts:
    if opt in ("-h","--help"):
        print "Command Overview:\n"
        print "\nSet the tweet-file for normalization:\n\tnormalize_tweets.py -i <file> | --inputfile <file>"
        print "\nEnable debug mode:\n\tnormalize_tweets.py -d\t      | --debug"
        print "\nCreate no normalization summary:\n\t normalize_tweets.py -n\t      | --noSummary"
        print "\nInclude normalization of abbreviations:\n\t normalize_tweets.py -f\t      | --full"
        print "\nShown unchanged tweets:\n\tnormalize_tweets.py -u\t      | --showUnchanged\t"
        print "\nShows this help:\n\tnormalize_tweets.py -h\t      | --help"

        print "\n\nDisables split string method:\n\tnormalize_tweets.py --noSplit\t"
        print "\nExample:\n\tnormalize_tweets.py -i myTweets.txt -fon --compare"
        sys.exit()
    elif opt in ("-i","--inputfile"):
        inputFileName = arg
        print "Processing: " + inputFileName
    elif opt in ("-c","--compare"):
        print "Info: Outputfile will be in compare mode"
        if not isFullCompareMode:
            isCompareMode = True
    elif opt in ("-o","--online"):
        print "Info: Online lookups will be done"
    elif opt in ("-n","--noSummary"):
        print "Info: No Summary will be created"
    elif opt == "--noSplit":
        print "Info: Split string method disabled"
    elif opt  in ("-u","--showUnchanged"):
        print "Info: Unchanged tweets will be shown"
        isShowUnchanged = True


    if opt in ("-d","--debug"):
        print "<< Debug Mode Active>>"
        debug = True

    if opt in ("-f","--full"):
        if  isCompareMode:
            isCompareMode = False
            isFullCompareMode = True
        isFullCompareMode = True
        isFullNorm = True
        print "Info: Normalization of abbreviations included"

## Read command line arguments


    ## Set booleans

if not argList:
    print "Info: No arguments submitted; processing with default values..."
    debug = False
    fileWrite  = True
    isCompareMode = False
    isShowUnchanged = True
    isCreateSummary = True
    enabledOnlineSearch = False
    enabledSpellingCorrection = True
    enabledSplitString = True

    # single debug flags
    showUsrN = False
    showHashTagN = False
    ## Set booleans

    inputFileName = "tweets_with_filtered_abb.txt"
    #inputFileName = "twita_tweets201202_01.txt"
    #inputFileName = "twita_tweets201305.txt"
    #inputFileName = "exampleTweets.txt"

if isCompareMode:
    if isShowUnchanged:
        outputFile = inputFileName + "_compare.cleantweets"
    else:
        outputFile = inputFileName + "_compareChanged.cleantweets"
else:
    outputFile = inputFileName + ".cleantweets"
outputFileSum = inputFileName + ".results"



##############################
###### help functions  #######
##############################

'''
Returns i-gram
'''
def getIGram(i,word):
    iGramList = []
    txtLength = len(word) + 1
    maxIgrams = txtLength - i if txtLength >= i else 0
    for n in range(maxIgrams):
        iGramList.append(word[n:n+i])
    #iGramList = [word[n:n+i] for n in range(maxIgrams)]
    return iGramList

'''
Returns a list with bigrams and skip-bigrams containing similiarty feature sets
of the input word.
'''
def getSmilarityFeatureSetFromWord(abb):
    if len(abb) < 3:
        return []
    similaritySet = []
    bigrams = getIGram(2, abb)
    startFeature = "$" + bigrams.pop(0)
    endFeature = bigrams.pop(-1) + "$"
    similaritySet.append(startFeature)
    similaritySet.append(endFeature)
    similaritySet.extend(set(bigrams))

    trigrams = getIGram(3, abb)
#     skipBigrams = []
#     for trigram in trigrams:
#         trigram = trigram[:1] + "|" + trigram[2:]
#         skipBigrams.append(trigram)
    skipBigrams = [trigram[:1] + "|" + trigram[2:] for trigram in trigrams]
    similaritySet.extend(set(skipBigrams))

    return similaritySet

'''
Returns the corrected version of the input word

1) filter all words in the vocab with levenshtein distance & ratio based on the input word
2) compute jaccardIndex of the input word and all remaining words in the vocab
3) return the word with the highest jaccardIndex
'''
def getCorrtectedWord(abb):
    w1similaritySet = getSmilarityFeatureSetFromWord(abb)
    highestPair = [0,""]
    hintersection = 0

    abb = abb.encode('utf-8')
    # Filter vocab to reduce for-loop overhead
    # Levenshtein Distance and ratio
    filteredVocab = [word2 if distance(abb, word2) < 3 and ratio(abb, word2) > 0.8 \
                     else "" for word2 in vocab]
    filteredVocab = filter(None, filteredVocab)

    for word2 in filteredVocab:

        jaccardIndex = 0

        #w2similaritySet = getSmilarityFeatureSetFromWord(word2) # Uncomment to create new Feature Sets
        w2similaritySet = vocab_FeatureSetDict[word2]


        # Fast jaccard | no method call to reduce time (no global vars used)
        intersection = len(set(w1similaritySet).intersection(w2similaritySet))
        if intersection > 4:
            if len(w1similaritySet) > 4:
                if w1similaritySet[0] in w2similaritySet:
                    intersection += 2
                if w1similaritySet[1] in w2similaritySet:
                    intersection += 1
            try:
                jaccardIndex = float(intersection) / len(set(w1similaritySet + w2similaritySet))
            except:
                jaccardIndex = 0
        # Fast jaccard | no method call to recude time

        if jaccardIndex > highestPair[0]:
            highestPair[0] = jaccardIndex
            highestPair[1] = word2
            hintersection = intersection
    wordLength = len(highestPair[1])
    if  wordLength < 5:
        if highestPair[0] > 0.37: #Standard metric 0.249  # 0.28
            if highestPair[1] not in abb:
                if debug: print "Word1: " + abb + " Cleaned: " + highestPair[1] + " intersection: " + str(hintersection)
                return highestPair[1]
            else:
                return abb
        else:
            return abb
    elif wordLength < 10:
        if highestPair[0] > 0.56: #Standard metric 0.42   # 0.44
            if highestPair[1] not in abb:
                if debug: print "Word2: " + abb + " Cleaned: " + highestPair[1] + " intersection: " + str(hintersection)
                return highestPair[1]
            else:
                return abb
        else:
            return abb
    else:
        if highestPair[0] > 0.6: #Standard metric  0.45
            if highestPair[1] not in abb:
                if debug: print "Word3: " + abb + " Cleaned: " + highestPair[1] + " intersection: " + str(hintersection)
                return highestPair[1]
            else:
                return abb
        else:
            return abb

def containsOnlyLetter(word):
    for char in word:
        if char not in lettersSet:
            return False
    return True

'''
Returns a list of splitted words from the input string
only if all substrings are words

'Thisgreatexample' -> 'This great example'
'Thisgretexample' -> ''
'''
def getInboundWords(abb):
    debug = False
    inboundWordList = []

    if len(abb) < 7:
        return []
    if abb in vocabSet:
        return []
    if not containsOnlyLetter(abb):
        return []

    if debug: print "check: " + abb

    isInbound = True


    while isInbound:
        wordList = [abb[:i+3] for i in xrange(len(abb)-2)]
        if debug and len(wordList) != 0: print wordList
        for word in reversed(wordList):
            if word in vocabSet:
                if debug: print "found word: " + word
                inboundWordList.append(word)
                abb = abb[len(word):]
                if debug: print "Compound: " + word + " " + abb


                isInbound = True
                break
            else:
                isInbound = False
        if not abb or len(wordList) == 0: isInbound = False # if abb == "" end loop

    if abb:
        return []
    else:
        return inboundWordList



### Read and save tweets
try:
    tweets = codecs.open(inputFileName, "r","utf-8").read().split("\n")
except:
    print "\nError cannot read input file: Make sure that the file '"+inputFileName+"' exists!"
    sys.exit()
if isFullNorm:
    outputFile = inputFileName + ".cleantweets"
tweets_normalizedOut = codecs.open(outputFile, "w","utf-8")
tweets_normalizedOut_Results = open(outputFileSum, "w")
### Read and save tweets

### Read other lists
twitterTopUsernameList = open("resources/twitter_usernames.txt").read().split("\n")
itBigramCharList = open("resources/itBigramCharList.txt").read().split("\n")
vocabList = open("resources/combinedVocab.vocab").read().split("\n")
vocab_FeatureSet = open("resources/combinedVocab_it_min5_featureset.vocab").read().split("\n")
it_stopwords = open("resources/stopwords_it_big.txt").read().split("\n")
### Read other lists


######################################
####### fill help lists/dicts  #######
######################################


## Fill twitter username dict with username and corresponding fullname
twitterTopUsernameDict = {}

for line in twitterTopUsernameList:
    elements = line.split("\t")
    username = elements[0][1:] # remove '@'
    fullname = elements[1]
    twitterTopUsernameDict[username.lower()] = fullname
## Fill twitter username dict with username and corresponding fullname


## Fill italian bigram char dict with char bigrams and probability
itBigramCharDict = {}

for line in itBigramCharList:
    line = line.split("\t")
    bigram = line[0].strip()
    bigramCount = line[1].strip()
    itBigramCharDict[bigram] = float(bigramCount)



## Fill vocab dict with words and frequency
vocabSet = {line.split("\t")[0] for line in vocabList}



## Fill vocab FeatureSet Dict with words and frequency
vocab_FeatureSetDict = {entry.split("\t")[0]:entry.split("\t")[1].split(" ") for entry in vocab_FeatureSet}
vocab = [entry for entry in vocab_FeatureSetDict]





eosCharList = ["!","?"]

lettersSet = set(string.letters)


emoticonCharList = [":",";","\\","'","$","%","^","&","*","(",")","_","-","=","+","]","[","{","}","~","`",">","<","/"]
specialCasesList = ["^^","*o*","O_O","O_o","T^T","xD","XD","o.O","O.O","O.o",":P",":p",":D",":d","=D",":S",":s","v.v","V.V"]
specialCasesLCList = ["^^","*o*","o_o","t^t","xd","o.o",":p",":d","=d",":s","v.v","9_9","^.^","^_~","~_^","~_~","n_n","ô_o",
                      "o_ô",":o","-.-","^_^","*hug*","tt.tt",":3","t_t","t__t","ù.ù","u.u","u_u","d:","ç_ç","ùù","._.","oò"]



setupTimeEnd = time.time()
print "#######################\n######Setup time:######\n#####" + str(setupTimeEnd-setupTimeStart)  + "#####"
######################################
####### run through all tweets #######
######################################

runTimeStart = time.time()

def normalizeTweets(tweets):
    tweets_normalized = []
    usernameLookupDict = {}

    parser = HTMLParser()

    countMultipleCharN = 0
    countLaughN = 0
    countEOSN = 0
    countPunctN = 0
    countRemoveSmilies = 0
    countRemoveLinks = 0
    countTokenN = 0
    countHTMLN = 0
    countSCN = 0
    countNONWordStrings = 0
    countRmvTags = 0
    countUsrN = 0
    countHashTagN = 0
    countSC = 0
    countSpaceN = 0

    usrAbbDict = {} # only in development-mode for new corpus needed

    tweetCount = 0
    for tweet in tweets:
        tweetWordListNormalized = []
        try: # if tweets are tab-separated in inputfile
            tweet = tweet.split("\t\t")[1]
        except:
            pass
        oriTweet = tweet
        if debug : print tweet
        countHTMLN += tweet.count('&#')
        tweet = tweet.replace('&#10;', "") # Remove HTML paragraph
        tweet = ' '.join(tweet.strip().split()) # Remove whitespaces in the tweet
        tweet = parser.unescape(tweet) # Replace HTML Tags with Unicode Characters
        tweetWordList = tweet.split(" ")
        countWordsInTweet = len(tweetWordList)
        countWord = 0

        hasTag = True
        ###{ Remove twitter tags in the beginning and the end of the tweet
        while (hasTag):
            # if tweet only contains twitter tags or links remove the tweet
            if len(tweetWordList) == 0:
                break
            hasStartTag = True
            hasEndTag = True
            firstToken = tweetWordList[0]
            lastToken = tweetWordList[-1]
            isPunct = False # Do not delete tags which ends whith punct (why don't you switch to #volunia?)
            isLink = False
            if lastToken.endswith("!") or lastToken.endswith(".") or lastToken.endswith("?"):
                try:
                    isPunct = lastToken[-2] != "." # delete tag with ends with "..."
                except:
                    isPunct = True
            if lastToken.startswith("http") or lastToken.startswith("https"): # rmv twitter tag followed by link
                isLink = True
                countRemoveLinks += 1
            if firstToken.startswith("@") or firstToken.startswith("#") or firstToken.startswith("RT") \
                or firstToken.startswith("MT") or firstToken.startswith("OH") \
                or firstToken.startswith(".@"): # rmv twitter tags in the beginning

                tweetWordList.pop(0)
                countRmvTags += 1
            else:
                hasStartTag = False
            if (not isPunct  and (lastToken.startswith("@") or lastToken.startswith("#")  or lastToken == "via") and len(tweetWordList) != 0) or isLink:
                tweetWordList.pop(-1)
                countRmvTags += 1 if not isLink else 0
            else:
                hasEndTag = False
            hasTag = True if hasStartTag or hasEndTag else False
        ###} Remove twitter tags in the beginning and the end of the tweet

        ### check for self defined abbs
#
#         #tweet = "Test with italian abbreviations (abb) like c.to=conto can=cortese attezione or c.c.p.=conto corrente postale"
#         #tweetWordList = tweet.split(" ")
#
#         wordCount = 0
#         for word in tweetWordList:
#             abb = ""
#             fullform = ""
#
#             if "=" in word and not word.startswith("http"):
#
#                 if "." in word:
#                     pairs = word.split("=")
#                     abb = pairs[0]
#                     fullform = pairs[1]
#                     punctAbb = abb.split(".")
#                     punctAbbCount = len(punctAbb)
#
#                     ## With punct 'c.c.p.=conto corrente postale'
#                     if punctAbbCount > 1:
#                         countParts = 0
#                         if fullform.startswith(punctAbb[countParts]):
#                             countParts += 1
#                             try:
#                                 while(countParts < punctAbbCount-1):
#                                     #print tweetWordList[wordCount+countParts]
#                                     if tweetWordList[wordCount+countParts].startswith(punctAbb[countParts]):
#                                         fullform += " " + tweetWordList[wordCount+countParts]
#                                     countParts += 1
#                             except IndexError:
#                                 pass
#                     ## With punct 'c.c.p.=conto corrente postale'
#                 else:
#                     pairs = word.split("=")
#                     abb = min(pairs,key=len)
#                     pairs.remove(abb)
#                     fullform = pairs[0]
#                     abbLen = len(abb)
#
#
#                     if abbLen > 1:
#                         countParts = 0
#                         if abb[0] == fullform[0]:
#                             try:
#                                 while(countParts < abbLen-1):
#                                     if abb[countParts+1] == tweetWordList[wordCount+countParts+1][0]:
#                                         fullform += " " + tweetWordList[wordCount+countParts+1]
#                                     countParts += 1
#                             except IndexError:
#                                 pass
#
#             if word.startswith("(") and word.endswith(")"):
#                 abb = word[1:-1] # remove parenthesis
#                 prevWord = tweetWordList[wordCount-1]
#                 fullform = ""
#                 if prevWord.startswith(abb.lower()) or abb[0].lower() == prevWord[0].lower():
#                     fullform = prevWord
#             wordCount += 1
#
#             # POTENZA,PZ ; Roma,RM ; Milano,MI
#             if abb and fullform:
#                 print "Abb: " +  abb
#                 print "Fullform: " + fullform
#                 usrAbbDict[abb] = fullform
#
#         ### check for self defined abbs

        singleCharList = []

        preWord = ""
        for abb in tweetWordList: # iterate through the words in a tweet
            countWord += 1

            ## Join spaced single characters [P R I M A V E R A = PRIMAVERA]
            if len(abb) == 1 and abb.isalpha():
                    singleCharList.append(abb)
                    continue
            else:
                if len(singleCharList) > 2: # join if list has minimum 3 elements
                    abb = "".join(singleCharList)
                    #tweetWordListNormalized.append(" ".join(singleCharList))
                elif len(singleCharList) != 0: # join if list has less then 3 elements, but is not empty
                    tweetWordListNormalized.extend(singleCharList)
                singleCharList = []
            ## Join spaced single characters [P R I M A V E R A = PRIMAVERA]

            if abb in vocabSet: # No normalization for words in the vocabulary
                tweetWordListNormalized.append(abb)
                continue
            if abb.lower() in vocabSet: # No normalization for words in the vocabulary
                tweetWordListNormalized.append(abb)
                continue
            ###### OUT-OF-VOC-WORDS: >>>>

            isRemoved = False
            ####{ Hasttag and Username Normalization
            if abb.startswith("@"):

                if showUsrN: print "Origin @Username: " + abb

                if len(abb[1:]) < 2:
                    if debug or showUsrN: print "####REMOVED too short username: " + abb
                    countRmvTags += 1
                    continue


                try:
                    ##{ Replace @username with full name from twitter profile
                    # no method call to speed up the program

                    hasPunct = False
                    lastChar = abb[-1]

                    if lastChar in ["?","!",".",",",";",":"]:
                        hasPunct = True

                    #remove_punctuation_map = dict((ord(char), None) for char in string.punctuation)
                    cleanedUsername = abb.encode('utf-8').translate(string.maketrans("",""),string.punctuation) # Rmv all puncts
                    #cleanedUsername = abb.translate(string.maketrans("",""),string.punctuation) # Rmv all puncts
                    cleanedUsernameLookup = abb.rstrip(",:.!?=;") # slower than translate

                    specialUsername = cleanedUsernameLookup.lower()[1:] # username with '_'

                    if cleanedUsername.lower() in twitterTopUsernameDict:      # if username found in dict
                        name = twitterTopUsernameDict[cleanedUsername.lower()] # replace with known full name
                        if hasPunct:
                            name += lastChar
                        if debug or showUsrN: print "####Normalized offline username: " + abb + " to: " + name
                        countUsrN += 1
                        tweetWordListNormalized.append(name.decode("utf-8"))
                    elif specialUsername in twitterTopUsernameDict:      # if username found in dict
                        name = twitterTopUsernameDict[specialUsername] # replace with known full name
                        if hasPunct:
                            name += lastChar
                        if debug or showUsrN: print "####Normalized offline username: " + abb + " to: " + name
                        countUsrN += 1
                        tweetWordListNormalized.append(name.decode("utf-8"))
                    elif cleanedUsername.lower() in usernameLookupDict:        # if username found in session dict
                        name = usernameLookupDict[cleanedUsername.lower()]     # replace with known full name
                        if hasPunct:
                            name += lastChar
                        if debug or showUsrN: print "####Normalized session username: " + abb + " to: " + name
                        countUsrN += 1
                        tweetWordListNormalized.append(name.decode("utf-8"))
                    else:
                        if enabledOnlineSearch:
                            url = 'https://twitter.com/'+ cleanedUsernameLookup[1:]
                            req = urllib2.Request(url)
                            conection = urllib2.urlopen(req)
                            content = conection.read()
                            result = re.search("title>(.*?) \(", content)
                            name = result.group(1) if result else ""
                            name = parser.unescape(name)

                            name = name.replace("_"," ")
                            name = filter(lambda x: (x.isalpha() or x==" " or x=="&" or x=="-"), name).strip()
                            if len(name) > 1:
                                usernameLookupDict[cleanedUsername.lower()] = name # save username lookups in this session
                                if hasPunct:
                                    name += lastChar
                                if debug or showUsrN: print "####Resolved username: "  + abb + " to Full name: " + name
                                abb = name
                                countUsrN += 1
                                tweetWordListNormalized.append(abb.decode("utf-8"))
                        else:
                            listOfUpperWords = re.findall('[A-Z][a-z]+', cleanedUsername)
                            if len(listOfUpperWords) > 0:
                                name = " ".join(listOfUpperWords)
                                if hasPunct:
                                    name += lastChar
                                tweetWordListNormalized.append(name.decode("utf-8"))
                                countUsrN += 1
                                if debug or showUsrN: print "####Normalized username: "  + abb + " to Full name: " + name
                            else:
                                isRemoved = True
                                countRmvTags += 1
                                if debug or showUsrN: print "####REMOVED online search for usernames disabled: " + abb
                    ##} Replace @username with full name from twitter profile
                except urllib2.HTTPError:
                    isRemoved = True
                    countRmvTags += 1
                    if debug or showUsrN: print "####REMOVED invalid username: " + abb
            elif abb.startswith("#"):
                isPrevPunct = False
                try:
                    prevWordLastChar = tweetWordList[countWord-2][-1]
                    if prevWordLastChar in [".","!","?"]:
                        isPrevPunct = True
                except:
                    isPrevPunct = True
                    pass

                ##{ Normalization of #hashtags
                if showHashTagN: print "Origin #Hashtag: " + abb
                hashTagText = abb[1:]
                if len(hashTagText) < 2:
                    if debug or showHashTagN: print "####REMOVED too short hashtag: " + abb
                    countRmvTags += 1
                    continue

                preStopword = ""
                if preWord and preWord in it_stopwords:
                    preStopword = preWord


                hasPunct = False
                hasImpPunct = False
                lastChar = hashTagText[-1]
                if lastChar in {"?","!",".",",",";",":"}:
                    if lastChar in {"!","?"}:
                        hasImpPunct = True
                    hasPunct = True

                remove_punctuation_map = dict((ord(char), None) for char in string.punctuation)
                #hashTagText = hashTagText.encode('utf-8').translate(remove_punctuation_map) # Rmv all puncts
                hashTagText = hashTagText.encode('utf-8').translate(string.maketrans("",""),string.punctuation) # Rmv all puncts
                hashTagText = hashTagText.decode("utf-8")
                if enabledSpellingCorrection:
                    hashTagText = hashTagText.lower()
                    correctedWord = getCorrtectedWord(hashTagText)
                    if correctedWord != hashTagText:
                        countSC += 1
                        if debug: print "NORMALIZED # spelling correction: " + hashTagText + " to: " + correctedWord
                        hashTagText = correctedWord
                inboundWords = getInboundWords(hashTagText)
                if len(inboundWords) > 0:
                    inboundWords = " ".join(inboundWords)
                    if hasPunct:                  # check if last char is punct
                        inboundWords += lastChar  # to keep sentence punctuation
                    if isPrevPunct:
                        inboundWords = inboundWords[0].upper() + inboundWords[1:] # First letter upper after sent-punct
                    tweetWordListNormalized.append(inboundWords)
                    continue
                listOfUpperWords = re.findall('[A-Z][a-z]+', hashTagText)

                if len(listOfUpperWords) > 0:
                    upperWords = " ".join(listOfUpperWords)
                    if hasPunct:                              # check if last char is punct
                        upperWords += lastChar                # to keep sentence punctuation
                    if debug or showHashTagN: print "####Normalized hashtag: " + abb + " to: " + upperWords
                    countHashTagN += 1
                    tweetWordListNormalized.append(upperWords)
                else:
                    if preStopword:
                        if hasPunct:
                            hashTagText += lastChar
                        if debug or showHashTagN: print "####Preword Stopword Hashtag: " + hashTagText \
                        + "with stopword: " + preStopword
                        countHashTagN += 1
                        tweetWordListNormalized.append(hashTagText)
                    if any(char.isdigit() for char in hashTagText): # check if hashtag contains numbers
                        isRemoved = True
                        countRmvTags += 1
                        if debug or showHashTagN: print "####REMOVED hashtag contains number: " + abb
                    else:
                        # keep if hashtag is valid word
                        if hashTagText in vocabSet:
                            if debug or showHashTagN: print "####Normalized hashtag word: " + abb \
                            + " to: " + hashTagText
                            countHashTagN += 1
                            if hasPunct:
                                hashTagText += lastChar
                            tweetWordListNormalized.append(hashTagText)
                        elif len(hashTagText) < 6:
                            if hasPunct:                               # check if last char is punct
                                hashTagText += lastChar                # to keep sentence punctuation
                            if debug or showHashTagN: print "####(Un)normalized hashtag: " + abb + " to: " + hashTagText
                            countHashTagN += 1
                            tweetWordListNormalized.append(hashTagText)
                        else:
                            if hasImpPunct: # Always keep hashtags before ! or ?
                                hashTagText += lastChar
                                countHashTagN += 1
                                tweetWordListNormalized.append(hashTagText)
                                if debug or showHashTagN: print "####Normalized hashtag!?: " + abb + " to: " + hashTagText
                            else:
                                if debug or showHashTagN: print "####REMOVED too long hashtag: " + abb
                                countRmvTags += 1
                                continue
                ##} Normalization of #hashtags

            ####} Hasttag and Username Normalization
            else:
                #### eosChar-Normalization (?+!+ == ?  | !+ == ! )
                if "?" or "!" or "," in abb: # Speed-up Version: seprate if's (only count if match)
                    qmCount = abb.count("?")
                    emCount = abb.count("!")
                    cmCount = abb.count(",")
                    if qmCount + emCount > 1:
                        if qmCount > 0:
                            countEOSN += 1
                            abb = re.sub(r"[?!]+", "?", abb)
                            #print "#####Replace####: " + abb
                            #tweetWordListNormalized.append(abb)
                        else:
                            countEOSN +=1
                            abb = re.sub(r"!+", "!", abb)
                            #tweetWordListNormalized.append(abb)
                            #print "#####Replace####: " + abb
                    if  cmCount > 1:
                        countPunctN += 1
                        abb = re.sub(r",+", ",", abb)
                    unspacedSeq = re.search(r"[A-Za-z]+([?!,])+[A-Za-z]+", abb)
                    if unspacedSeq:
                        countPunctN += 1
                        punct = unspacedSeq.group(1)
                        abb = abb.replace(punct, punct + " ") # Capitalize after punct (!/?/.)
                    #charSeqs = re.split("!|,|\.|\?",abb)
                #### eosChar-Normalization

                ###### Smiley, Link and other unwanted Tokens-Normalization
                countEChars = 0
                for emoticonChar in emoticonCharList:
                    if emoticonChar in abb:
                        countEChars += 1
                if countEChars > 1:
                    if debug: print "####Suggest-to-REMOVE: " + abb
                    if len(abb) > 3:

                        if abb.startswith("http:") or abb.startswith("https:"): ######### Remove Links
                            if debug: print "####REMOVED link: " + abb
                            countRemoveLinks += 1
                            isRemoved = True
                        else:
                            #abb = re.sub(r"[:;\\'$%^&*()_=+][{}~`></-]+","",abb)
                            #abb = re.sub(r"[\W_]+","",abb)

                            ### Keep Words sticked together with Smilies etc. but remove Smilies etc.
                            pos = 0
                            lastPos = 0
                            nonAlpha = False

                            for char in abb:
                                pos += 1
                                if char.isalpha():
                                    nonAlpha = False
                                    lastPos = pos
                                elif "'" in char and not nonAlpha:
                                    nonAlpha = True
                                    lastPos = pos
                                elif "?" in char or "!" in char and not nonAlpha:
                                    lastPos = pos
                                    break
                                else:
                                    break

                            if lastPos > 1:
                                abb = abb[:lastPos]
                                if debug:  print "####NOTREMOVE: "+ abb
                                countTokenN += 1
                                #tweetWordListNormalized.append(abb)
                            ### Keep Words sticked together with Smilies etc. but remove Smilies etc.
                            else:
                                ### Keep "bracked words" e.g.: '(Reuters)'
                                isBrackedWord = False
                                if abb.startswith("(") and abb.endswith(")"):
                                    for char in abb[1:-1]:
                                        if char.isalpha():
                                            isBrackedWord = True
                                        else:
                                            isBrackedWord = False
                                            break
                                if isBrackedWord:
                                    if debug: print "####NOTREMOVE: "+ abb
                                    countTokenN += 1
                                    #tweetWordListNormalized.append(abb)
                                ### Keep "bracked words" e.g.: '(Reuters)'
                                else:
                                    countRemoveSmilies += 1
                                    isRemoved = True
                                    if debug: print "####4REMOVED: " + abb
                    else:
                        countRemoveSmilies += 1
                        if debug: print "####REMOVED: " + abb
                        isRemoved = True
                    ###### Smiley, Link and other unwanted Tokens-Normalization
                else:
                    # Maximal 1 emoticonChar
                    ###### Multiple Char Normalization
                    #if "@" in abb or "#" in abb: # or "." in abb:
                    #    pass
                    #else:
                    if "..." in abb: # DO NOT normalize "..." but "......"
                        dotCount = abb.count(".")
                        if dotCount > 3:
                            countMultipleCharN += 1
                            oldWord = abb
                            abb = re.sub(r"\.\.\.+","...", abb)
                            if debug: print "####Normalized ...: " + abb + " Word: " + oldWord
                    else:
                        ## Trimming leading double letters "ggrazie"
                        if len(abb) > 4 and abb[0] == abb[1] and abb[-1] != ".":
                            if abb[2] != abb[1] and not abb[0].isdigit() \
                               and abb[0].isalpha() and abb[2].isalpha() and abb[3].isalpha() and abb[4].isalpha(): # Three ident letters will be trimmed later
                                #countMultipleCharN += 1
                                abbOld = abb
                                abbCheck = abb[1:]
                                countUpperCaseLetter = sum(1 for c in abb if c.isupper())
                                if countUpperCaseLetter > 1: # Normalize Uppercase only
                                    if abbCheck.lower() in vocabSet: # if exists in vocab
                                        abb = abbCheck
                                        if debug: print "####TrimmedUpper: " + abb + " Word: " + abbOld
                                else:
                                    abb = abbCheck
                                    if debug: print "####Trimmed: " + abb + " Word: " + abbOld

                                # Trimming leading double letters on a row "aahh"
                                if abb[1] == abb[2] and not abb[1].isdigit() and abb[1].isalpha():
                                    abbOld = abb
                                    abb = abb[:2] + abb[3:]
                                    if debug: print "####Trimmed2: " + abb + " Word: " + abbOld
                                # Trimming leading double letters  on a row "aahh"
                        ## Trimming leading double letters "ggrazie"
                        multiple = ""
                        lastChar = ""
                        savedMultiple = []
                        for char in abb + " ":
                            if lastChar in char:
                                multiple += char
                            else:
                                if len(multiple) > 2:
                                    savedMultiple.append(multiple)
                                multiple = char
                            lastChar = char
                        if len(savedMultiple) > 0:
                            countMultipleCharN += 1
                            oldWord = abb
                            for multiple in savedMultiple:
                                abb = abb.replace(multiple,multiple[0])

                            if debug: print "####Normalized: " + abb + " Word: " + oldWord

                    if abb not in vocabSet:
                        if containsOnlyLetter(abb) and not abb[0].isupper():
                            correctedWord = getCorrtectedWord(abb)
                            if correctedWord != abb:
                                countSC += 1
                                if debug: print "NORMALIZED spelling correction1: " + abb + \
                                " to: " + correctedWord
                                abb = correctedWord
                    ## Laugh-Normalization
                    laughSeq = abb.lower()
                    count1 = laughSeq.count("ha")
                    count2 = laughSeq.count("ah")
                    if count1 > 1 or count2 > 1:
                        oldWord = abb
                        charSeqs = re.split("!|,|\.|\?",abb)
                        for charSeq in charSeqs: # ("hahaha!!!Bella => haha!!!Bella")
                            charSeqOld = charSeq
                            charSeq = charSeq.lower()
                            count1 = charSeq.count("ha")
                            count2 = charSeq.count("ah")
                            if count1 > 1 or count2 > 1:
                                oldWord = abb
                                countLaughN += 1
                                abb = abb.replace(charSeqOld,"haha")
                                if debug: print "####L-Normalized: " +  abb + " normalized: " + oldWord
                    ## Laugh-Normalization

                    blaSeq = laughSeq # to lower
                    count1 = blaSeq.count("bla")
                    if count1 > 2:
                        print abb
                    ###### Multiple Char Normalization
                if not isRemoved:
                    #### Special Cases
                    for seq in specialCasesLCList:
                        if seq.decode("UTF-8") in abb.lower():
                            oldWord = abb
                            countSCN += 1
                            #seq = seq.replace("^","\^")
                            #seq = seq.replace(".","\.")
                            #seq = seq.replace("*","\*")
                            abb = re.sub("(?i)"+re.escape(seq), "", abb)
                            if debug: print "####Normalized-SC: " + abb + " Word: " + oldWord

                    ## Normalize strings containing *
                    if "*" in abb:

                        countSCN += 1 # Rounded
                        abb = re.sub("\*.?\*","",abb) # Removes Smilies like *v*| *o*

                        if len(abb) == 4:
                            abb = re.sub(r"\*(.)?\1\*","",abb) # Removes *cc* / *xx* / *..* ...
                        abb = abb.lstrip("*") # Strip leading *'s
                        abb = abb.rstrip("*") # Strip trailing *'s
                    ## Normalize strings containing *
                    abb = abb.replace('""',"") # Remove leftover "|'-Chars
                    abb = abb.replace("''","") # Remove leftover "|'-Chars

                    ## Tokens like: 'djdjdnjdjdbdjdbhsjsjsjsjsjjs' or 'uhuhuhyuhuhuhu' or 'eheheheh'
                    # Ungewöhnliche hohe Anzahl an Zeichen insgesamt zu Anzahl an verschiedenen Buchstaben
                    notRemove = True
                    nonWord = abb.lower()

                    try:
                        int(nonWord)
                    except:
                        containsNumber = False
                        for char in nonWord:
                            if char.isdigit():
                                containsNumber = True
                        if not containsNumber:
                            countCharList = []

                            for char in string.ascii_lowercase:
                                countChar = nonWord.count(char)
                                if countChar > 0:
                                    countCharList.append(countChar)
                            if len(countCharList) > 0 and len(nonWord) > 7 and \
                               len(nonWord) / len(countCharList) > 3:
                                countSCN += 1
                                notRemove = False
                                if debug: print "####REMOVED: " + abb

                    ## Tokens like: 'djdjdnjdjdbdjdbhsjsjsjsjsjjs' or 'uhuhuhyuhuhuhu' or 'eheheheh'

                    ## Unwanted UTF8-Chars
                    if '😉'.decode("UTF-8") in abb:
                        countSCN += 1
                        if debug: print "####STRIPPED:" + abb
                        abb = abb.replace("😓😉".decode("UTF-8"),u"")
                        abb = abb.replace("💙😉".decode("UTF-8"),u"")
                        abb = abb.replace("😜😉".decode("UTF-8"),u"")
                        abb = abb.replace("😉".decode("UTF-8"),u"")
                    ## Unwanted UTF8-Chars

                    ## Remove abnormal strings "Djfrfcsdgsdyhcjh" / "IOSODIJSAJDOSAD"
                    #if len(abb) > 10: # String length > 10 and contains only letters
                    abbLength = len(abb)
                    if containsOnlyLetter(abb) and abbLength > 1:
                        bigramList = getIGram(2, abb.lower())
                        if abbLength > 10:
                            totalSum  = 0
                            for bigramm in bigramList:
                                try:
                                    totalSum += itBigramCharDict[bigramm]
                                except:
                                    pass # totalSum += 0
                            ratio = totalSum / len(abb)
                            if ratio < 0.003:
                                if debug: print "####REMOVED: " + abb + " Ratio: " + str(ratio)
                                countNONWordStrings += 1
                                notRemove = False
                        else:
                            d = max({x:bigramList.count(x) for x in bigramList}.values())
                            if d > 2:
                                if debug: print "####REMOVED 3x+ same bigram: " + abb
                                countNONWordStrings += 1
                                notRemove = False

                    ## Remove abnormal strings "Djfrfcsdgsdyhcjh" / "IOSODIJSAJDOSAD"


                    ## Spelling correction
                    if enabledSpellingCorrection and notRemove:
                        if len(abb) > 4:
                            if abb not in vocabSet:
                                #print abb
                                if containsOnlyLetter(abb) and not abb[0].isupper():
                                    correctedWord = getCorrtectedWord(abb)
                                    if correctedWord != abb:
                                        countSC += 1
                                        if debug: print "NORMALIZED spelling correction: " + abb + \
                                        " to: " + correctedWord
                                        abb = correctedWord
                    ## Spelling correction


                    #### Special Cases
                    if notRemove and abb:

                        ### 'Splitstringintowords' -> 'Split string into words'
                        splittedWords = getInboundWords(abb)
                        if enabledSplitString and len(splittedWords) > 0:
                            if debug: print "#############Space Insert: " + abb + " => "+ str(splittedWords)
                            countSpaceN += 1
                            tweetWordListNormalized.extend(splittedWords)
                        ### 'Splitstringintowords' -> 'Split string into words'
                        else:
                            tweetWordListNormalized.append(abb)
            preWord = abb
        tweet_normalized = u" ".join(tweetWordListNormalized)
        if tweet in tweet_normalized:
            if fileWrite:
                tweetCount += 1
                normFlagList.append(tweetCount)
                if isShowUnchanged:
                    if isCompareMode:
                        tweets_normalizedOut.write(oriTweet + "\nUnchanged####: " + tweet_normalized + "\n\n")
                    else:
                        tweets_normalizedOut.write(tweet_normalized + "\n")
                else:
                    tweets_normalizedOut.write(tweet_normalized + "\n")
        else:
            if fileWrite:
                tweetCount += 1
                normFlagList.append(tweetCount)
                if isCompareMode:
                    tweets_normalizedOut.write(oriTweet + "\n" + tweet_normalized + "\n\n")
                else:
                    tweets_normalizedOut.write(tweet_normalized + "\n")
            if debug: print "NORMALIZED: " + tweet_normalized
        tweets_normalized.append(tweet_normalized)




    result = "\n\nNormalization Results: '"+ inputFileName + "'\n" + \
            "\nMultiple Char Normalization: " + str(countMultipleCharN) + \
             "\nEndofSentence Normalization: " + str(countEOSN) + \
             "\nPunctuation Normalization: " + str(countPunctN) + \
             "\nNoisy abb Normalization: " + str(countTokenN) + \
             "\nLaugh Normalization: " + str(countLaughN) + \
             "\nHTML Normalization: " + str(countHTMLN) + \
             "\nSpecial Cases Normalization: " + str(countSCN) + \
             "\nRemove Smiley and unwanted tokens: " +str(countRemoveSmilies) + \
             "\nRemove URLs: " + str(countRemoveLinks) + \
             "\nRemove non-word strings containing only letters: " + str(countNONWordStrings) + \
             "\nTwitter @username Normalization: " + str(countUsrN) + \
             "\nTwitter #hashtag Normalization: " + str(countHashTagN) + \
             "\nRemove twitter tags: " + str(countRmvTags) + \
             "\nSpelling Correction Normalization: " + str(countSC) + \
             "\nSplitted String into words (missing Space) Normalization: " + str(countSpaceN) + \
             "\n\nTotal Normalizations: " + str(countMultipleCharN+countEOSN+countPunctN+countTokenN \
                                                +countRemoveSmilies+countRemoveLinks+countHTMLN+countSCN \
                                                +countNONWordStrings+countUsrN+countHashTagN+countRmvTags+countSC) + \
            "\nTweet Count: " + str(len(tweets))


    print result

    if isCreateSummary:
        tweets_normalizedOut_Results.write(result)


normalizeTweets(tweets)



runTimeEnd = time.time()

print "\n\n######################\n#######Runtime:#######\n#####"+ str(runTimeEnd - runTimeStart)+"####"


if isFullNorm:
    arguments = "-i " + outputFile
    if isFullCompareMode:
        arguments += " -c"
    print "Info: calling abbreviations script"

    os.system("python normalize_abb_tweets_PAISA_Model.py " + arguments)
    #subprocess.Popen("python normalize_abb_tweets_PAISA_Model.py " + arguments, shell=True)

    compareFileAbb = open(outputFile + "_compare.normtweets").read().split("\n")
    fullOutputCompare = open(outputFile + "_full_compare.normtweets","w")
    countTweet = 0
    offset = 1
    #import random
    #choosen = random.sample(normFlagList,100)
    for numOfNormedTweet in normFlagList:
        #if numOfNormedTweet in choosen:
        try: # if tweets are tab-separated in inputfile
            if "\t" in tweets[numOfNormedTweet-1]:
                tweet = tweets[numOfNormedTweet-1].split("\t\t")[1]
            else:
                tweet = tweets[numOfNormedTweet-1]
        except:
            pass
        try:
            fullOutputCompare.write(tweet + "\n" + compareFileAbb[offset] + "\n\n")
        except:
            pass
        if offset == 1:
            offset = 4
        else:
            offset += 3


